import io
import math
from ks3.encryption import Crypts

SEEK_SET = getattr(io, 'SEEK_SET', 0)
SEEK_CUR = getattr(io, 'SEEK_CUR', 1)
SEEK_END = getattr(io, 'SEEK_END', 2)

class EncryptFp (object):
    """
    A class will return you a fp inside which data are encrypted.
    """
    def __init__(self, fp, crypt_context, type, isUploadFirstPart=False, isUploadLastPart=False):
        self.fp = fp
        self.first_iv = crypt_context.first_iv
        self.calc_iv = ""
        self.crypt_context = crypt_context
        self.crypt_handler = Crypts(crypt_context.key)
        self.type = type
        self.block_size = 16
        self.isUploadFirstPart = isUploadFirstPart
        self.isUploadLastPart = isUploadLastPart
        self.seek_pos = SEEK_SET
        self.block_count = 0
        self.block_total_count = self.get_total_count(fp)

    def get_total_count(self, fp):
        fp.seek(0, SEEK_END)
        count = math.ceil(float(fp.tell())/8192)
        fp.seek(0, SEEK_SET)
        return count

    def __getattr__(self, name):
        func = getattr(self.__dict__['fp'], name)
        if callable(func):
            if name == "read":
                def my_wrapper(*args, **kwargs):
                    data = func(*args, **kwargs)
                    self.block_count += 1
                    if len(data) == 0:
                        return None
                    # print len(data)
                    if self.type == "put":              
                        if self.block_count == 1:
                            if self.block_total_count == 1:
                                encrypt_data = self.crypt_handler.encrypt(data,self.first_iv)
                            else:
                                encrypt_data = self.crypt_handler.encrypt_without_padding(data,self.first_iv)
                            self.calc_iv = encrypt_data[-self.block_size:]
                            self.crypt_context.calc_iv = self.calc_iv
                            encrypt_data = self.first_iv+encrypt_data
                        elif self.block_count == self.block_total_count:
                            encrypt_data = self.crypt_handler.encrypt(data,self.calc_iv)
                        else:
                            encrypt_data = self.crypt_handler.encrypt_without_padding(data, self.calc_iv)
                            self.calc_iv = encrypt_data[-self.block_size:]
                            self.crypt_context.calc_iv = self.calc_iv
                    elif self.type == "upload_part":
                        need_prefix = False
                        if self.isUploadFirstPart and self.block_count == 1:
                            pre_iv = self.first_iv
                            need_prefix = True
                        else:
                            last_part_num = self.crypt_context.part_num - 1
                            if last_part_num > 0 and self.block_count == 1:
                                if self.crypt_context.iv_dict.get(last_part_num):
                                    self.calc_iv = self.crypt_context.iv_dict.get(last_part_num)
                            else:
                                if not self.calc_iv:
                                    raise ValueError(
                                        "upload part[%d] encryption error:calc_vi miss" % self.crypt_context.part_num)
                            pre_iv = self.calc_iv
                        if self.isUploadLastPart and self.block_count == self.block_total_count:
                            encrypt_data = self.crypt_handler.encrypt(data, pre_iv)
                        else:
                            encrypt_data = self.crypt_handler.encrypt_without_padding(data, pre_iv)
                        if need_prefix:
                            encrypt_data = self.first_iv + encrypt_data
                        self.calc_iv = encrypt_data[-self.block_size:]
                        self.crypt_context.iv_dict[self.crypt_context.part_num] = self.calc_iv
                        # if self.isUploadFirstPart:
                        #     # For multi, the first part's first part will add a prefix of iv.
                        #     if not self.calc_iv:
                        #         if self.isUploadLastPart and self.block_count == self.block_total_count:
                        #             # A very special circumstance: a short piece of data that is both the first of the
                        #             # first and the last of the last.
                        #             encrypt_data = self.crypt_handler.encrypt(data, self.first_iv)
                        #             encrypt_data = self.first_iv + encrypt_data
                        #         else:
                        #             encrypt_data = self.crypt_handler.encrypt_without_padding(data,self.first_iv)
                        #             encrypt_data = self.first_iv+encrypt_data
                        #     elif not self.isUploadLastPart:
                        #         encrypt_data = self.crypt_handler.encrypt_without_padding(data, self.calc_iv)
                        #     elif self.isUploadLastPart and self.block_count == self.block_total_count:
                        #         # When the part is the firstPart AND the lastPart.
                        #         encrypt_data = self.crypt_handler.encrypt(data, self.calc_iv)
                        #     elif self.isUploadLastPart:
                        #         encrypt_data = self.crypt_handler.encrypt_without_padding(data, self.calc_iv)
                        #     else:
                        #         raise Exception
                        # elif not self.isUploadLastPart:
                        #     # The normal part that is neither the first nor the last one.
                        #     if self.block_count == 1:
                        #         self.calc_iv = self.crypt_context.iv_dict[self.crypt_context.part_num-1]
                        #     encrypt_data = self.crypt_handler.encrypt_without_padding(data, self.calc_iv)
                        # else:
                        #     # The last part.
                        #     # The last part's parts use 'encrypt' instead of 'encrypt_without_padding'
                        #     # because the last part's last part need paddling.
                        #     if self.block_count == 1 and self.block_count != self.block_total_count:
                        #         encrypt_data = self.crypt_handler.encrypt_without_padding(data, self.crypt_context.iv_dict[self.crypt_context.part_num-1])
                        #     elif self.block_count == 1 and self.block_count == self.block_total_count:
                        #         encrypt_data = self.crypt_handler.encrypt(data, self.crypt_context.iv_dict[self.crypt_context.part_num-1])
                        #     elif self.block_count == self.block_total_count:
                        #         encrypt_data = self.crypt_handler.encrypt(data, self.calc_iv)
                        #     else:
                        #         encrypt_data = self.crypt_handler.encrypt_without_padding(data,self.calc_iv)
                        # self.calc_iv = encrypt_data[-self.block_size:]
                        # self.crypt_context.iv_dict[self.crypt_context.part_num] = self.calc_iv
                        # print len(encrypt_data), self.block_count, self.block_total_count
                    return encrypt_data
            if name == "seek":
                def my_wrapper(*args, **kwargs):
                    ret = func(*args, **kwargs)
                    # self.seek_pos = args[1]
                    return ret
            if name == "tell":
                def my_wrapper(*args, **kwargs):
                    ret = func(*args, **kwargs)
                    if self.type == "upload_part":
                        if self.isUploadFirstPart:
                            return ret + 16
                        else:
                            return ret
                    return ret+16

            return my_wrapper
        else:
            return func

    def __len__(self):
        self.seek(0,SEEK_END)
        length = self.tell()
        self.seek(0,SEEK_SET)
        blocksize = 16
        if self.type == "put" or (self.type=="upload_part" and self.isUploadLastPart==True):
            pad = blocksize - length % blocksize
            return length+pad
        else:
            return length



# -*- coding: utf-8 -*-
"""
KS3 SDK 流上传示例
演示如何使用新添加的 set_contents_from_stream 和 set_contents_from_stream_simple 方法
"""

import os
from io import BytesIO
from ks3.connection import Connection
from ks3.bucket import Bucket

# 配置信息 - 请根据实际情况修改
AK = os.getenv('KS3_IAM_ACCESS_KEY_ID', '<YOUR_ACCESS_KEY>')
SK = os.getenv('KS3_IAM_ACCESS_KEY_SECRET', '<YOUR_SECRET_KEY>')
BUCKET_NAME = os.getenv('KS3_BUCKET', '<YOUR_BUCKET_NAME>')
ENDPOINT = 'ks3-cn-shanghai.ksyuncs.com'

def test_stream_upload():
    """测试流上传功能"""
    # 初始化连接
    conn = Connection(AK, SK, host=ENDPOINT, is_secure=True)

    # 获取存储桶
    bucket = conn.get_bucket(BUCKET_NAME)
    print("1. 使用完整参数版本上传...")
    memory_stream2 = BytesIO(b"Hello, KS3 Stream Upload! This is a test with full parameters.")
    key3 = bucket.new_key('stream_test/full_params_stream.txt')
    result3 = key3.set_contents_from_stream(
        stream=memory_stream2,
        headers={'Content-Type': 'text/plain'},
        replace=True,
        policy='private'
    )
    print(f"   上传结果: status={result3.status}, request_id={result3.response_metadata.request_id}")

    # # 2. 带进度回调的上传
    # print("2. 带进度回调的上传...")
    # def progress_callback(sent, total):
    #     if total > 0:
    #         print(f"   上传进度: {sent}/{total} bytes ({sent/total*100:.1f}%)")
    #
    # large_data = BytesIO(b'A' * 1024 * 100)  # 100KB的数据
    # key4 = bucket.new_key('stream_test/large_data_stream.txt')
    # result4 = key4.set_contents_from_stream(
    #     stream=large_data,
    #     cb=progress_callback,
    #     num_cb=5
    # )
    # print(f"   上传结果: status={result4.status}, request_id={result4.response_metadata.request_id}")

    # 清理临时文件
    try:
        os.remove('/tmp/test_stream_upload.txt')
    except:
        pass

    print("\n=== 所有测试完成 ===")

if __name__ == '__main__':
    test_stream_upload()


from ks3.responseResult import ResponseMetadata


class VpcAccessBlock:
    def __init__(self, rules=None, *args, **kwargs):
        if rules is None:
            rules = []
        self.rules = rules

        self.response_metadata = ResponseMetadata(**kwargs)

    def startElement(self, name, attrs, connection):
        if name == 'Rule':
            self.rules.append(Rule())
            return self.rules[-1]
        return None

    def endElement(self, name, value, connection):
        setattr(self, name, value)

    def to_xml(self):
        s = u'<VpcAccessBlockConfiguration>'
        for i in self.rules:
            s += i.to_xml()
        s += '</VpcAccessBlockConfiguration>'
        return s


class Rule:
    def __init__(self, rule_id=None, region=None, vpc_ids=None, allow_access_buckets=None, enabled=None):
        if vpc_ids is None:
            vpc_ids = []
        if allow_access_buckets is None:
            allow_access_buckets = []

        self.rule_id = rule_id
        self.region = region
        self.vpc_ids = vpc_ids
        self.allow_access_buckets = allow_access_buckets
        self.enabled = enabled

    def startElement(self, name, attrs, connection):
        return None

    def endElement(self, name, value, connection):
        if name == 'RuleID':
            self.rule_id = value
        elif name == 'Region':
            self.region = value
        elif name == 'ID':
            self.vpc_ids.append(value)
        elif name == 'Name':
            self.allow_access_buckets.append(value)
        elif name == 'Status':
            self.enabled = True if value.lower() == 'enabled' else False
        else:
            setattr(self, name, value)

    def to_xml(self):
        s = u'<Rule>'
        if self.rule_id is not None:
            s += '<RuleID>{0}</RuleID>'.format(self.rule_id)
        if self.region is not None:
            s += '<Region>{0}</Region>'.format(self.region)
        if self.vpc_ids is not None and len(self.vpc_ids) > 0:
            s += '<VPC>'
            for i in self.vpc_ids:
                s += '<ID>{0}</ID>'.format(i)
            s += '</VPC>'
        if self.allow_access_buckets is not None and len(self.allow_access_buckets) > 0:
            s += '<BucketAllowAccess>'
            for i in self.allow_access_buckets:
                s += '<Name>{0}</Name>'.format(i)
            s += '</BucketAllowAccess>'
        if self.enabled is not None:
            s += '<Status>{0}</Status>'.format('Enabled' if self.enabled else 'Disabled')
        s += '</Rule>'
        return s


# coding:utf-8
"""
批量任务相关的 XML 解析器类

支持以下操作：
- KS3PutObjectDataRedundancyTransition: 批量转换为同城冗余
- KS3RestoreObject: 批量解冻
- KS3PutObjectAcl: 批量修改ACL
- KS3DeleteObject: 批量删除

相关辅助类：
- KS3AccessControlList: AccessControlList 结构
- KS3Grant: Grant 结构（包含 Grantee 和 Permission）
"""

from ks3.responseResult import ResponseMetadata


class KS3PutObjectDataRedundancyTransition(object):
    """
    批量转换为同城冗余操作参数
    """
    def __init__(self, data_redundancy_type=None):
        self.data_redundancy_type = data_redundancy_type

    def to_xml(self):
        s = '<KS3PutObjectDataRedundancyTransition>\n'
        if self.data_redundancy_type is not None:
            s += '    <DataRedundancyType>%s</DataRedundancyType>\n' % self.data_redundancy_type
        s += '</KS3PutObjectDataRedundancyTransition>\n'
        return s

    def startElement(self, name, attrs, connection):
        return None

    def endElement(self, name, value, connection):
        if name == 'DataRedundancyType':
            self.data_redundancy_type = value


class KS3RestoreObject(object):
    """
    批量解冻操作参数
    允许设置 days、tier 和 storage_class 参数
    storage_class 表示需要解冻的存储类型，可选值：ARCHIVE、COLD_ARCHIVE
    其中 days、storage_class 为必选参数，tier 为可选参数（仅在冷归档类型时可用，必选）
    """
    def __init__(self, days=None, tier=None, storage_class=None):
        self.days = days
        self.tier = tier
        self.storage_class = storage_class

    def to_xml(self):
        s = '<KS3RestoreObject>\n'
        if self.days is not None:
            s += '    <Days>%s</Days>\n' % self.days
        if self.tier is not None:
            s += '    <Tier>%s</Tier>\n' % self.tier
        if self.storage_class is not None:
            s += '    <StorageClass>%s</StorageClass>\n' % self.storage_class
        s += '</KS3RestoreObject>\n'
        return s

    def startElement(self, name, attrs, connection):
        return None

    def endElement(self, name, value, connection):
        if name == 'Days':
            self.days = value
        elif name == 'Tier':
            self.tier = value
        elif name == 'StorageClass':
            self.storage_class = value


class KS3PutObjectAcl(object):
    """
    批量修改ACL操作参数

    支持两种方式：
    1. CannedAccessControlList: 预设ACL ('private' 或 'public-read')
    2. AccessControlList: 自定义ACL，包含多个Grant

    示例1 (CannedAccessControlList):
        KS3PutObjectAcl('private') 或 KS3PutObjectAcl('public-read')

    示例2 (AccessControlList with Grants):
        grants = [
            KS3Grant('12345678', 'FULL_CONTROL'),
            KS3Grant('87654321', 'READ')
        ]
        KS3PutObjectAcl(access_control_list=KS3AccessControlList(grants))
    """
    def __init__(self, canned_acl=None, access_control_list=None):
        self.canned_acl = canned_acl
        self.access_control_list = access_control_list

    def to_xml(self):
        s = '<KS3PutObjectAcl>\n'
        if self.canned_acl is not None:
            s += '    <CannedAccessControlList>%s</CannedAccessControlList>\n' % self.canned_acl
        if self.access_control_list is not None:
            s += '    <AccessControlList>\n'
            for grant in self.access_control_list.grants:
                s += '        <Grant>\n'
                s += '            <Grantee>%s</Grantee>\n' % grant.grantee
                s += '            <Permission>%s</Permission>\n' % grant.permission
                s += '        </Grant>\n'
            s += '    </AccessControlList>\n'
        s += '</KS3PutObjectAcl>\n'
        return s

    def startElement(self, name, attrs, connection):
        if name == 'AccessControlList':
            self.access_control_list = KS3AccessControlList()
            return self.access_control_list
        elif name == 'Grant':
            if self.access_control_list is None:
                self.access_control_list = KS3AccessControlList()
            grant = KS3Grant()
            self.access_control_list.grants.append(grant)
            return grant
        return None

    def endElement(self, name, value, connection):
        if name == 'CannedAccessControlList':
            self.canned_acl = value


class KS3AccessControlList(object):
    """
    AccessControlList for KS3PutObjectAcl
    """
    def __init__(self, grants=None):
        self.grants = grants if grants is not None else []

    def add_grant(self, grant):
        self.grants.append(grant)

    def startElement(self, name, attrs, connection):
        if name == 'Grant':
            grant = KS3Grant()
            self.grants.append(grant)
            return grant
        return None

    def endElement(self, name, value, connection):
        pass


class KS3Grant(object):
    """
    Grant for KS3PutObjectAcl
    包含 Grantee 和 Permission
    """
    def __init__(self, grantee=None, permission=None):
        self.grantee = grantee
        self.permission = permission

    def to_xml(self):
        s = '<Grant>\n'
        s += '    <Grantee>%s</Grantee>\n' % self.grantee
        s += '    <Permission>%s</Permission>\n' % self.permission
        s += '</Grant>\n'
        return s

    def startElement(self, name, attrs, connection):
        return None

    def endElement(self, name, value, connection):
        if name == 'Grantee':
            self.grantee = value
        elif name == 'Permission':
            self.permission = value


class KS3DeleteObject(object):
    """
    批量删除操作参数
    """
    def __init__(self):
        pass

    def to_xml(self):
        return '<KS3DeleteObject>\n</KS3DeleteObject>\n'

    def startElement(self, name, attrs, connection):
        return None

    def endElement(self, name, value, connection):
        pass


class JobOperation(object):
    """
    任务操作
    支持多种操作类型同时设置：
    - KS3RestoreObject: 批量解冻
    - KS3PutObjectAcl: 批量修改ACL
    - KS3DeleteObject: 批量删除
    - KS3PutObjectDataRedundancyTransition: 批量转为同城冗余
    """
    def __init__(self, data_redundancy_transition=None, restore_object=None,
                 put_object_acl=None, delete_object=None):
        self.data_redundancy_transition = data_redundancy_transition
        self.restore_object = restore_object
        self.put_object_acl = put_object_acl
        self.delete_object = delete_object

    def to_xml(self):
        s = '<Operation>\n'
        if self.data_redundancy_transition is not None:
            s += '    %s' % self.data_redundancy_transition.to_xml()
        if self.restore_object is not None:
            s += '    %s' % self.restore_object.to_xml()
        if self.put_object_acl is not None:
            s += '    %s' % self.put_object_acl.to_xml()
        if self.delete_object is not None:
            s += '    %s' % self.delete_object.to_xml()
        s += '</Operation>\n'
        return s

    def startElement(self, name, attrs, connection):
        if name == 'KS3PutObjectDataRedundancyTransition':
            self.data_redundancy_transition = KS3PutObjectDataRedundancyTransition()
            return self.data_redundancy_transition
        elif name == 'KS3RestoreObject':
            self.restore_object = KS3RestoreObject()
            return self.restore_object
        elif name == 'KS3PutObjectAcl':
            self.put_object_acl = KS3PutObjectAcl()
            return self.put_object_acl
        elif name == 'KS3DeleteObject':
            self.delete_object = KS3DeleteObject()
            return self.delete_object
        return None

    def endElement(self, name, value, connection):
        pass


class JobManifestFilter(object):
    """
    任务清单过滤器
    包含 Bucket 和多个 Prefix
    """
    def __init__(self, bucket=None, prefixes=None):
        self.bucket = bucket
        self.prefixes = prefixes if prefixes is not None else []

    def add_prefix(self, prefix):
        """添加前缀"""
        if prefix is not None:
            self.prefixes.append(prefix)

    def to_xml(self):
        s = '<Filter>\n'
        if self.bucket is not None:
            s += '    <Bucket>%s</Bucket>\n' % self.bucket
        for prefix in self.prefixes:
            s += '    <Prefix>%s</Prefix>\n' % prefix
        s += '</Filter>\n'
        return s

    def startElement(self, name, attrs, connection):
        return None

    def endElement(self, name, value, connection):
        if name == 'Bucket':
            self.bucket = value
        elif name == 'Prefix':
            self.prefixes.append(value)


class JobManifestLocation(object):
    """
    任务清单位置
    包含多个 Filter
    """
    def __init__(self, filters=None):
        self.filters = filters if filters is not None else []

    def add_filter(self, filter_obj):
        """添加过滤器"""
        if filter_obj is not None:
            self.filters.append(filter_obj)

    def to_xml(self):
        s = '<Location>\n'
        for f in self.filters:
            s += '    %s' % f.to_xml()
        s += '</Location>\n'
        return s

    def startElement(self, name, attrs, connection):
        if name == 'Filter':
            new_filter = JobManifestFilter()
            self.filters.append(new_filter)
            return new_filter
        return None

    def endElement(self, name, value, connection):
        pass


class JobManifestSpec(object):
    """
    任务清单规格
    """
    def __init__(self, format=None):
        self.format = format

    def to_xml(self):
        s = '<Spec>\n'
        if self.format is not None:
            s += '    <Format>%s</Format>\n' % self.format
        s += '</Spec>\n'
        return s

    def startElement(self, name, attrs, connection):
        return None

    def endElement(self, name, value, connection):
        if name == 'Format':
            self.format = value


class JobManifest(object):
    """
    任务清单
    包含 Location 和 Spec 字段
    """
    def __init__(self, location=None, spec=None):
        self.location = location
        self.spec = spec

    def to_xml(self):
        s = '<Manifest>\n'
        if self.location is not None:
            s += '    %s' % self.location.to_xml()
        if self.spec is not None:
            s += '    %s' % self.spec.to_xml()
        s += '</Manifest>\n'
        return s

    def startElement(self, name, attrs, connection):
        if name == 'Location':
            self.location = JobManifestLocation()
            return self.location
        elif name == 'Spec':
            self.spec = JobManifestSpec()
            return self.spec
        return None

    def endElement(self, name, value, connection):
        pass


class CreateJobRequest(object):
    """
    创建任务请求
    """
    def __init__(self, manifest=None, operation=None, priority=None, description=None,
                 report=None, client_request_token=None):
        self.manifest = manifest
        self.operation = operation
        self.priority = priority
        self.description = description
        self.report = report
        self.client_request_token = client_request_token

    def to_xml(self):
        s = '<CreateJobRequest>\n'
        if self.manifest is not None:
            s += '    %s' % self.manifest.to_xml()
        if self.operation is not None:
            s += '    %s' % self.operation.to_xml()
        if self.priority is not None:
            s += '    <Priority>%s</Priority>\n' % self.priority
        if self.description is not None:
            s += '    <Description>%s</Description>\n' % self.description
        if self.report is not None:
            s += '    %s' % self.report.to_xml()
        if self.client_request_token is not None:
            s += '    <ClientRequestToken>%s</ClientRequestToken>\n' % self.client_request_token
        s += '</CreateJobRequest>\n'
        return s

    def startElement(self, name, attrs, connection):
        if name == 'Manifest':
            self.manifest = JobManifest()
            return self.manifest
        elif name == 'Operation':
            self.operation = JobOperation()
            return self.operation
        elif name == 'Report':
            self.report = JobReport()
            return self.report
        return None

    def endElement(self, name, value, connection):
        if name == 'Priority':
            self.priority = value
        elif name == 'Description':
            self.description = value
        elif name == 'ClientRequestToken':
            self.client_request_token = value


class JobReport(object):
    """
    任务报告配置
    """
    def __init__(self, enabled=None, bucket=None, prefix=None, report_scope=None):
        self.enabled = enabled
        self.bucket = bucket
        self.prefix = prefix
        self.report_scope = report_scope

    def to_xml(self):
        s = '<Report>\n'
        if self.enabled is not None:
            s += '    <Enabled>%s</Enabled>\n' % ('true' if self.enabled else 'false')
        if self.bucket is not None:
            s += '    <Bucket>%s</Bucket>\n' % self.bucket
        if self.prefix is not None:
            s += '    <Prefix>%s</Prefix>\n' % self.prefix
        if self.report_scope is not None:
            s += '    <ReportScope>%s</ReportScope>\n' % self.report_scope
        s += '</Report>\n'
        return s

    def startElement(self, name, attrs, connection):
        return None

    def endElement(self, name, value, connection):
        if name == 'Enabled':
            self.enabled = value.lower() == 'true'
        elif name == 'Bucket':
            self.bucket = value
        elif name == 'Prefix':
            self.prefix = value
        elif name == 'ReportScope':
            self.report_scope = value


class CreateJobResult(object):
    """
    创建任务响应
    """
    def __init__(self, job_id=None, *args, **kwargs):
        self.job_id = job_id
        self.response_metadata = ResponseMetadata(**kwargs)

    def startElement(self, name, attrs, connection):
        return None

    def endElement(self, name, value, connection):
        if name == 'JobId':
            self.job_id = value


class JobProgressSummary(object):
    """
    任务进度摘要
    """
    def __init__(self, total_number_of_tasks=None, number_of_tasks_succeeded=None,
                 number_of_tasks_failed=None, *args, **kwargs):
        self.total_number_of_tasks = total_number_of_tasks
        self.number_of_tasks_succeeded = number_of_tasks_succeeded
        self.number_of_tasks_failed = number_of_tasks_failed
        self.response_metadata = ResponseMetadata(**kwargs)

    def startElement(self, name, attrs, connection):
        return None

    def endElement(self, name, value, connection):
        if name == 'TotalNumberOfTasks':
            self.total_number_of_tasks = value
        elif name == 'NumberOfTasksSucceeded':
            self.number_of_tasks_succeeded = value
        elif name == 'NumberOfTasksFailed':
            self.number_of_tasks_failed = value


class DescribeJobResult(object):
    """
    查询任务详情响应
    """
    def __init__(self, job_id=None, status=None, creation_time=None,
                 priority=None, description=None, manifest=None, operation=None,
                 progress_summary=None, report=None, termination_date=None,
                 *args, **kwargs):
        self.job_id = job_id
        self.status = status
        self.creation_time = creation_time
        self.priority = priority
        self.description = description
        self.manifest = manifest
        self.operation = operation
        self.progress_summary = progress_summary
        self.report = report
        self.termination_date = termination_date
        self.response_metadata = ResponseMetadata(**kwargs)

    def startElement(self, name, attrs, connection):
        if name == 'Manifest':
            self.manifest = JobManifest()
            return self.manifest
        elif name == 'Operation':
            self.operation = JobOperation()
            return self.operation
        elif name == 'ProgressSummary':
            self.progress_summary = JobProgressSummary()
            return self.progress_summary
        elif name == 'Report':
            self.report = JobReport()
            return self.report
        return None

    def endElement(self, name, value, connection):
        if name == 'JobId':
            self.job_id = value
        elif name == 'Status':
            self.status = value
        elif name == 'CreationTime':
            self.creation_time = value
        elif name == 'Priority':
            self.priority = value
        elif name == 'Description':
            self.description = value
        elif name == 'TerminationDate':
            self.termination_date = value


class JobListMember(object):
    """
    任务列表中的单个任务信息
    DescribeJobResult 的简化版，不包含 manifest、report 和 response_metadata
    """
    def __init__(self, job_id=None, status=None, creation_time=None,
                 priority=None, description=None, operation=None,
                 termination_date=None):
        self.job_id = job_id
        self.status = status
        self.creation_time = creation_time
        self.priority = priority
        self.description = description
        self.operation = operation
        self.termination_date = termination_date

    def startElement(self, name, attrs, connection):
        if name == 'Operation':
            self.operation = JobOperation()
            return self.operation
        return None

    def endElement(self, name, value, connection):
        if name == 'JobId':
            self.job_id = value
        elif name == 'Status':
            self.status = value
        elif name == 'CreationTime':
            self.creation_time = value
        elif name == 'Priority':
            self.priority = value
        elif name == 'Description':
            self.description = value
        elif name == 'TerminationDate':
            self.termination_date = value


class ListJobsResult(object):
    """
    列出任务响应
    """
    def __init__(self, jobs=None, next_token=None, *args, **kwargs):
        self.jobs = jobs if jobs is not None else []
        self.next_token = next_token
        self.response_metadata = ResponseMetadata(**kwargs)

    def startElement(self, name, attrs, connection):
        if name == 'Member':
            job = JobListMember()
            self.jobs.append(job)
            return job
        return None

    def endElement(self, name, value, connection):
        if name == 'NextToken':
            self.next_token = value


class UpdateJobPriorityResult(object):
    """
    更新任务优先级响应
    """
    def __init__(self, job_id=None, priority=None, *args, **kwargs):
        self.job_id = job_id
        self.priority = priority
        self.response_metadata = ResponseMetadata(**kwargs)

    def startElement(self, name, attrs, connection):
        return None

    def endElement(self, name, value, connection):
        if name == 'JobId':
            self.job_id = value
        elif name == 'Priority':
            self.priority = int(value) if value else None


from ks3.responseResult import ResponseMetadata


class BucketTransferAcceleration:

    def __init__(self, enabled=None, *args, **kwargs):
        self.enabled = 'true' if enabled else 'false'

        self.response_metadata = ResponseMetadata(**kwargs)

    def startElement(self, name, attrs, connection):
        return None

    def endElement(self, name, value, connection):
        if name == 'Enabled':
            self.enabled = value
        setattr(self, name, value)

    def to_xml(self):
        s = u'<TransferAccelerationConfiguration>'
        if self.enabled is not None:
            s += '<Enabled>%s</Enabled>' % self.enabled
        s += '</TransferAccelerationConfiguration>'
        return s


# coding=utf-8
from ks3.connection import Connection

####################公共头#########################

# 金山云主账号 AccessKey 拥有所有API的访问权限，风险很高。强烈建议您创建并使用子账号账号进行 API 访问或日常运维，请登录 https://uc.console.ksyun.com/pro/iam/#/user/list 创建子账号。
# 通过指定 host(Endpoint)，您可以在指定的地域创建新的存储空间。host(Endpoint) 以北京为例，其它 Region 请按实际情况填写。
conn = Connection('<yourAccessKeyId>', '<yourAccessKeySecret>', host='ks3-cn-beijing.ksyuncs.com')
# 获取存储空间实例
b = conn.get_bucket('<yourBucketName>')

#################### 内容 #########################

#################### 下载 #########################

# 下载 Object，并且作为字符串返回
# 填写Object完整路径。Object完整路径中不能包含Bucket名称。
k = b.get_key('<yourKeyName>')
# 获取对象内容，并解码
# 对于1000字节大小的文件，正常的下载范围取值为0~999。
# 获取0~999字节范围内的数据，包括0和999，共1000个字节的数据。如果指定的范围无效（比如开始或结束位置的指定值为负数，或指定值大于文件大小），则下载整个文件。
# 如示例，通过『范围下载』下载 2Byte 数据
s = k.get_contents_as_string(byte_range=(0, 1)).decode()
print(s)

# 流式下载 Object
# 适用于文件过大、下载时间过长的情况，可以通过流式下载，分批处理，直到完成
# 填写Object完整路径。Object完整路径中不能包含Bucket名称。
k = b.get_key('<yourKeyName>')
# 按照字节大小读取，比如300
bytes = k.read(300)
while bytes:
    s = bytes.decode()
    # 对内容进行处理，比如打印
    print(s)
    bytes = k.read(300)

# 下载 Object，并且保存到文件中
# 填写Object完整路径。Object完整路径中不能包含Bucket名称。
k = b.get_key('<yourKeyName>')
# 填写文件路径，保存到文件
# 对于1000字节大小的文件，正常的下载范围取值为0~999。
# 获取0~999字节范围内的数据，包括0和999，共1000个字节的数据。如果指定的范围无效（比如开始或结束位置的指定值为负数，或指定值大于文件大小），则下载整个文件。
# 如示例，通过『范围下载』下载 2Byte 数据
k.get_contents_to_filename("<savedFilePath>", byte_range=(0, 1))

# 生成下载外链地址
# 对私密属性的文件生成下载外链（该链接具有时效性）
# 填写Object完整路径。Object完整路径中不能包含Bucket名称。
k = b.get_key('<yourKeyName>')
if k:
    # 下载外链地址 60s 后过期
    # image_attrs 为图像的指令或者样式参数字符串
    url = k.generate_url(60, image_attrs='@base@tag=imgScale&w=500')
    # 下载外链地址在时间点 1492073594 后过期，1492073594 为 Unix Time
    # image_attrs 为图像的指令或者样式参数字符串
    k.generate_url(1492073594, expires_in_absolute=True, image_attrs='@base@tag=imgScale&w=500')

#################### 上传 #########################

# 将指定目录下某一个文件上传，同时可以指定文件 ACL
# 填写Object完整路径。Object完整路径中不能包含Bucket名称。
k = b.new_key('<yourKeyName>')
# x-kss-storage-class有效值为"STANDARD"、"STANDARD_IA"。"STANDARD"表示标准存储，"STANDARD_IA"表示低频存储，如果不指定，默认为标准存储。
headers = {"x-kss-storage-class": "STANDARD_IA"}
# 填写本地文件的完整路径。
# object policy : 'private' or 'public-read'
ret = k.set_contents_from_filename("<yourSourceFilePath>", policy="private", headers=headers)
# 请求ID。请求ID是本次请求的唯一标识，强烈建议在程序日志中添加此参数。
print(ret.headers['x-kss-request-id'])
# ETag是put_object方法返回值特有的属性，用于标识一个Object的内容。
print(eval(ret.headers['ETag']))
# HTTP返回码。
if ret and ret.status == 200:
    print("上传成功")

# 将字符串作为 value 上传
# 填写Object完整路径。Object完整路径中不能包含Bucket名称。
k = b.new_key('<yourKeyName>')
# 参数为文件内容
k.set_contents_from_string('<yourFileContent>', headers=None)

# 将字符串作为 value 上传，并设置对象 tag
# key 和 value 需要 url 编码
taggingStr = '<key>=<value>'
headers = {'x-kss-tagging': taggingStr}
ret = k.set_contents_from_string('<yourFileContent>', headers=headers)

# 上传成功后，向回调地址POST数据
# 需要设置`x-kss-callbackurl`和`x-kss-callbackbody`请求头，详见文档[上传回调处理](https://docs.ksyun.com/documents/956)。
# x-kss-callbackurl 为发起回调时请求的服务器地址
# x-kss-callbackbody 为发起回调时请求的body的值
headers = {"x-kss-callbackurl": "<yourCallbackUrl>", "x-kss-callbackbody": "objectKey=${key}&etag=${etag}&uid=123"}
ret = k.set_contents_from_filename("<yourSourceFilePath>", headers=headers)

# 使用签名URL临时授权上传文件
# 填写Object完整路径。Object完整路径中不能包含Bucket名称。
k = b.new_key("<yourKeyName>")
if k:
    # 生成上传文件的签名URL，有效时间为60秒。
    url = k.get_presigned_url(60)
    print(url)

#################### 管理文件 #########################

# 获取文件元信息（大小、最后更新时间等）
# 填写Object完整路径。Object完整路径中不能包含Bucket名称。
k = b.get_key('<yourKeyName>')
if k:
    print(k.name, k.size, k.last_modified)

# 获取并打印出object的ACL信息（暂时无法说明）
# 填写Object完整路径。Object完整路径中不能包含Bucket名称。
# policy = b.get_acl("<yourKeyName>")
# print(policy.to_xml())

# 设置object的ACL
# object的ACL可选择 'private' or 'public-read'，下面示例设置为'public-read'
# 填写Object完整路径。Object完整路径中不能包含Bucket名称。
b.set_acl("public-read", "<yourKeyName>")

# 添加存储类型header，此处更改文件存储类型为归档类型
# x-kss-storage-class有效值为"STANDARD"、"STANDARD_IA"。"STANDARD"表示标准存储，"STANDARD_IA"表示低频存储，如果不指定，默认为标准存储。
headers = {'x-kss-storage-class': 'STANDARD_IA'}
# 更改文件存储类型。通过复制并添加新header的方式实现，故方法名为copy_key
# <yourDstKeyName> 为新对象名称，<yourSrcBucketName>是原存储空间名称，<yourSrcKeyName>是原对象名称
b.copy_key('<yourDstKeyName>', '<yourSrcBucketName>', '<yourSrcKeyName>', headers=headers)

# 添加元数据相关header，此处更改content-type和元数据指令
# 详情参考[PUT Object Copy](https://docs.ksyun.com/documents/941)
headers = {'content-type': 'text/plain', 'x-kss-metadata-directive': 'REPLACE'}
# 更改文件元数据信息。通过复制并添加新header的方式实现，故方法名为copy_key
# <yourDstKeyName> 为新对象名称，<yourSrcBucketName>是原存储空间名称，<yourSrcKeyName>是原对象名称
b.copy_key('<yourDstKeyName>', '<yourSrcBucketName>', '<yourSrcKeyName>', headers=headers)

# 列举文件
#### 列举 Bucket 内的文件或者目录

from ks3.connection import Connection
from ks3.prefix import Prefix
from ks3.key import Key

# 金山云主账号 AccessKey 拥有所有API的访问权限，风险很高。强烈建议您创建并使用子账号账号进行 API 访问或日常运维，请登录 https://uc.console.ksyun.com/pro/iam/#/user/list 创建子账号。
# 通过指定 host(Endpoint)，您可以在指定的地域创建新的存储空间。host(Endpoint) 以北京为例，其它 Region 请按实际情况填写。
conn = Connection('<yourAccessKeyId>', '<yourAccessKeySecret>', host='ks3-cn-beijing.ksyuncs.com')
# 获取存储空间实例
b = conn.get_bucket('<yourBucketName>')

keys = b.list()
for k in keys:
    if isinstance(k, Key):
        print('文件:%s' % k.name)
    elif isinstance(k, Prefix):
        print('目录:%s' % k.name)

# 列举 Bucket 内指定前缀的文件
# 列举 images 文件夹下的所有文件。比如b.list(prefix="images/")
keys = b.list(prefix="<yourPrefix>")

# 列举 Bucket 内以指定分隔符分组的文件
# delimiter 为对文件名称进行分组的字符；delimiter 为空时，默认为'/'
# 返回结果中 delimiter 分隔符之前的字符会放入 commonPrefixes 中，可以类比理解为文件夹
# 只列举该文件夹下的文件和子文件夹（目录）名称，子文件夹下的文件和文件夹不显示。
keys = b.list(delimiter='/')

# 列举指定字符串（object名称）之后（字典排序）的所有文件。同名 object 也会被排除。
keys = b.list(marker='<yourMarker>')

# 列举指定个数的文件。如下列举10个文件
keys = b.list(max_keys=10)

# 列举 Bucket 内指定前缀的文件以及指定时间区间的文件
# start_time = 1625460400 这个时间之前的数据
# end_time   = 1625460400 这个时间之后的数据
# start_time=1640331446, end_time=1641895096 这个时间中间的数据
keys = b.listObjects(prefix="local/load", start_time=1640331446, end_time=1641895096)


# 列举目录下的文件大小
# 计算指定目录下的文件大小
def calcFolderSize(bucket, folder):
    length = 0
    keys = bucket.list(prefix=folder)
    for k in keys:
        if isinstance(k, Key):
            length += k.size
    return length


from ks3.prefix import Prefix


def getFolderSizeInBucket():
    keys = b.list(delimiter='/')
    for k in keys:
        # 判断是否为目录
        if isinstance(k, Prefix):
            print('dir: ' + k.name + '  size:' + str(calcFolderSize(b, k.name)) + "Byte")


# 删除文件。<yourObjectName>表示删除文件时需要指定包含文件后缀在内的完整路径。如 images/test.jpg
# 暂不支持删除文件夹
b.delete_key('<yourKeyName>')

# 拷贝文件
# 空间名称必须有效，同时用户需要拥有对拷贝对象的读权限。
# <yourDstKeyName> 为新对象名称，<yourSrcBucketName>是原存储空间名称，<yourSrcKeyName>是原对象名称
b.copy_key('<yourDstKeyName>', '<yourSrcBucketName>', '<yourSrcKeyName>')

# 解冻文件
# 对归档Object进行解冻
k = b.get_key('<yourKeyName>')
k.restore_object()

# 设置对象数据冗余类型
# 切换对象的数据冗余类型，支持：LRS（本地冗余）、ZRS（同城冗余）
if k:
    # 将对象冗余类型切换为LRS（本地冗余）
    k.set_object_data_redundancy_transition(data_redundancy_type='LRS')
    print(f"对象 {k.name} 已切换为LRS（本地冗余）")

# 抓取网络资源上传
# 从第三方URL拉取文件，并上传至KS3某个 bucket 中存储成名为 object 的文件。
k = b.new_key('www-logo')
k.fetch_object(source_url='http://fe.ksyun.com/project/resource/img/www-logo.png')
# 参见：[金山云官方文档-Put Object Fetch](https://docs.ksyun.com/documents/949)


# 对象标签
# 获取对象标签
k = b.get_key('<yourKeyName>')
tagging = k.get_object_tagging()
print(tagging.to_xml())

# 设置对象标签
from ks3.tagging import Tag

k = b.get_key('<yourKeyName>')
tagging = [Tag('<key>', '<value>')]
k.set_object_tagging(tagging)

# 删除对象标签
k = b.get_key('<yourKeyName>')
k.delete_object_tagging()

# 字符串追加上传，仅作特性功能展示，其他参考常规上传
# 填写Object完整路径。Object完整路径中不能包含Bucket名称。
k = b.new_key('<yourKeyName>')
# 第一个参数为文件内容；position为插入位置
ret = k.append_object_from_string('<yourFileContent1>', position='<appendPosition>')
if ret and ret.status == 200:
    print("追加上传成功")
# 如有需要，获取下一次append的位置
print("x-kss-next-append-position:", ret.headers['x-kss-next-append-position'])
# 继续追加
ret = k.append_object_from_string('<yourFileContent2>', position=int(ret.headers['x-kss-next-append-position']))
if ret and ret.status == 200:
    print("继续追加上传成功")

# 文件追加上传，用法参考"字符串追加上传"
# 填写Object完整路径。Object完整路径中不能包含Bucket名称。
k = b.new_key('<yourKeyName>')
# 第一个参数为文件路径；position为插入位置
ret = k.append_object_from_filename("<yourSourceFilePath>", position='<appendPosition>')
if ret and ret.status == 200:
    print("文件追加上传成功")

# 异步上传文件，以asyncio为例
import asyncio

k = b.new_key('<yourKeyName>')
ret = asyncio.run(k.upload_file_async("<yourSourceFilePath>"))
if ret and ret.status == 200:
    print("上传成功")

# 异步下载文件，以asyncio为例
import asyncio

k = b.get_key('<yourKeyName>')
asyncio.run(k.download_file_async("<yourSourceFilePath>"))



# coding=utf-8
import math, os
from ks3.connection import Connection
from filechunkio import FileChunkIO

####################公共头#########################
# 金山云主账号 AccessKey 拥有所有API的访问权限,风险很高。强烈建议您创建并使用子账号账号进行 API 访问或日常运维,请登录 https://uc.console.ksyun.com/pro/iam/#/user/list 创建子账号。
# 通过指定 host(Endpoint),您可以在指定的地域创建新的存储空间。host(Endpoint) 以北京为例,其它 Region 请按实际情况填写。
conn = Connection('<yourAccessKeyId>', '<yourAccessKeySecret>', host='ks3-cn-beijing.ksyuncs.com')
# 获取存储空间实例
b = conn.get_bucket('<yourBucketName>')
# 源文件路径
source_path = '<yourSourceFilePath>'
# 源文件大小
source_size = os.stat(source_path).st_size
# 对象键名
key_name = '<yourKeyName>'

# 初始化分片。获取初始化的 uploadId,之后的操作中将会用到
# 如需在初始化分片时设置文件存储类型,请在 initiate_multipart_upload 中设置相关 headers
# x-kss-storage-class有效值为"STANDARD"、"STANDARD_IA"。"STANDARD"表示标准存储,"STANDARD_IA"表示低频存储,如果不指定,默认为标准存储。
headers = {"x-kss-storage-class": "STANDARD"}
mp = b.initiate_multipart_upload(key_name, headers=headers)
upload_id = mp.id

# 举例以 50 MiB 为分片大小
chunk_size = 52428800
chunk_count = int(math.ceil(source_size*1.0 / chunk_size*1.0))

# 通过 FileChunkIO 将文件分片并逐个上传
for i in range(chunk_count):
    offset = chunk_size * i
    bytes = min(chunk_size, source_size - offset)
    with FileChunkIO(source_path, 'r', offset=offset, bytes=bytes) as fp:
        # 直接调用 bucket 的 upload_part_from_file 方法上传分片
        response = b.upload_part_from_file(
            key_name=key_name,
            upload_id=upload_id,
            fp=fp,
            part_num=i + 1
        )
        print('上传分片 %d, 状态码: %s' % (i + 1, response.status))

# 列出指定上传任务中所有已上传的分片信息
# part_number_marker 指定应该从哪个分片开始列举,只有比设定值大的分片才会被列举
result = b.list_parts(
    bucket_name=b.name,
    key_name=key_name,
    upload_id=upload_id,
    part_number_marker=0
)
print('已上传的分片:')
for part in result.data:
    print('part_number:%s, size:%s, etag:%s' % (part.part_number, part.size, part.etag))

# 方式一：传统方式 - 手动生成完成上传的 XML 请求体
xml_body = '<CompleteMultipartUpload>\n'
for part in result.data:
    xml_body += '  <Part>\n'
    xml_body += '    <PartNumber>%d</PartNumber>\n' % part.part_number
    xml_body += '    <ETag>%s</ETag>\n' % part.etag
    xml_body += '  </Part>\n'
xml_body += '</CompleteMultipartUpload>'

# 完成分片上传（使用 xml_body 参数）
ret = b.complete_multipart_upload(key_name, upload_id, xml_body=xml_body)
if ret and ret.status == 200:
    print("上传成功（方式一）")

# 方式二：新方式 - 直接传入 Part 列表
# 注意：实际使用时，每次只能完成一次上传。以下代码仅作为示例展示不同调用方式
# 可以直接传入 list_parts 返回的 IterableResponseResult
# ret2 = b.complete_multipart_upload(key_name, upload_id, parts=result)
# if ret2 and ret2.status == 200:
#     print("上传成功（方式二 - 使用 IterableResponseResult）")

# 方式三：使用 Part 对象列表
# parts_list = result.data  # 这是 Part 对象列表
# ret3 = b.complete_multipart_upload(key_name, upload_id, parts=parts_list)
# if ret3 and ret3.status == 200:
#     print("上传成功（方式三 - 使用 Part 对象列表）")

# 方式四：使用字典列表
# parts_dict = [
#     {'part_number': 1, 'etag': '"xxx"'},
#     {'part_number': 2, 'etag': '"yyy"'},
# ]
# ret4 = b.complete_multipart_upload(key_name, upload_id, parts=parts_dict)
# if ret4 and ret4.status == 200:
#     print("上传成功（方式四 - 使用字典列表）")

# 列举分片上传事件
# 列举 Bucket 中所有的分片上传事件
key_marker = '<yourKey>'
for p in b.list_multipart_uploads(key_marker=key_marker):
    print('uploadId:%s, key:%s' % (p.id, p.key_name))
    # 列出每个上传任务的分片
    parts_result = b.list_parts(
        bucket_name=b.name,
        key_name=p.key_name,
        upload_id=p.id
    )
    for part in parts_result.data:
        print('  part_number:%s, size:%s, etag:%s, last_modified:%s' % 
              (part.part_number, part.size, part.etag, part.last_modified))

# 取消指定upload_id的分片上传事件,已上传的分片会被删除
# 示例:取消某个上传任务
# response = b.cancel_multipart_upload(key_name, upload_id)
# print('取消上传响应:', response.status)


from ks3.responseResult import ResponseMetadata

"""
data_redundancy_type: 数据冗余类型，取值为LRS或ZRS
"""

class BucketDataRedundancySwitch:
    def __init__(self, data_redundancy_type=None, switch_time=None, *args, **kwargs):
        self.data_redundancy_type = data_redundancy_type
        self.switch_time = switch_time

        self.response_metadata = ResponseMetadata(**kwargs)

    def startElement(self, name, attrs, connection):
        return None

    def endElement(self, name, value, connection):
        if name == 'DataRedundancyType':
            self.data_redundancy_type = value
        if name == 'SwitchTime':
            self.switch_time = value
        setattr(self, name, value)

    def to_xml(self):
        s = u'<BucketDataRedundancySwitch>'
        if self.data_redundancy_type is not None:
            s += '<DataRedundancyType>%s</DataRedundancyType>' % self.data_redundancy_type
        if self.switch_time is not None:
            s += '<SwitchTime>%s</SwitchTime>' % self.switch_time
        s += '</BucketDataRedundancySwitch>'
        return s


from ks3.responseResult import ResponseMetadata

"""
<ServerSideEncryptionConfiguration>
    <Rule>
        <ApplyServerSideEncryptionByDefault>
            <SSEAlgorithm>AES256</SSEAlgorithm>
        </ApplyServerSideEncryptionByDefault>
    </Rule>
</ServerSideEncryptionConfiguration>
"""


class BucketEncryption(object):

    def __init__(self, algorithm=None, *args, **kwargs):
        self.algorithm = algorithm

        self.response_metadata = ResponseMetadata(**kwargs)

    def startElement(self, name, value, connection):
        return None

    def endElement(self, name, value, connection):
        if name == 'SSEAlgorithm':
            self.algorithm = value

    def to_xml(self):
        s = '<ServerSideEncryptionConfiguration>' \
            '<Rule>' \
            '<ApplyServerSideEncryptionByDefault>'

        if self.algorithm is not None:
            s += '<SSEAlgorithm>%s</SSEAlgorithm>' % self.algorithm
        s += '</ApplyServerSideEncryptionByDefault>' \
             '</Rule>' \
             '</ServerSideEncryptionConfiguration>'
        return s



# coding=utf-8
import math
import os
import time

import ks3.multipart
from ks3 import utils
from ks3.connection import Connection

ak = os.getenv('KS3_IAM_ACCESS_KEY_ID', '<YOUR_ACCESS_KEY>')
sk = os.getenv('KS3_IAM_ACCESS_KEY_SECRET', '<YOUR_SECRET_KEY>')
bucket_name = os.getenv('KS3_BUCKET', '<KS3_BUCKET>')
endpoint = os.getenv('KS3_ENDPOINT', 'ks3-cn-beijing-internal.ksyuncs.com')

key_name = 'test-file'
file_path = './test-file'

# 默认开启自动校验crc
# 关闭需要 conn.enable_crc = False 或 Connection(ak, sk, host=endpoint, enable_crc=False)
conn = Connection(ak, sk, host=endpoint)
bucket = conn.get_bucket(bucket_name)

# 下载对象后，计算服务端返回的crc64值，与本地读取计算的crc64值进行对比
def test_downloadObject_checkingCrc64():
    k = bucket.get_key(key_name)

    # 下载成功之后，会生成server_crc
    # 读取对象数据之后，会生成client_crc
    data = k.read()
    print(data)

    print(k.server_crc, k.client_crc)
    # 支持crc64功能之前的文件，server_crc为None
    if k.server_crc is None:
        print("server crc is None, maybe server not support crc64")
    else:
      assert k.server_crc == k.client_crc, f"k.server_crc: {k.server_crc}, k.client_crc: {k.client_crc}"

# 下载为字符串后，计算本地crc64值，与服务端crc64值进行对比
def test_downloadObjectToString_checkingCrc64():
    k = bucket.get_key(key_name)
    s = k.get_contents_as_string().decode()
    client_crc = utils.compute_data_crc64(s)
    # client_crc = k.client_crc
    # 支持crc64功能之前的文件，server_crc为None
    if k.server_crc is None:
      print("server crc is None, maybe server not support crc64")
    else:
      print("client_crc:", client_crc, "server_crc:", k.server_crc)
      assert client_crc == k.server_crc, f"client_crc: {client_crc}, server_crc: {k.server_crc}"


# 下载到文件，计算本地文件crc64值，与服务端crc64值进行对比
def test_downloadObjectToFile_checkingCrc64():
    k = bucket.get_key(key_name)
    k.get_contents_to_filename('downloaded-' + key_name)
    # client_crc = utils.compute_file_crc64('downloaded-' + key_name)
    client_crc = k.client_crc
    # 支持crc64功能之前的文件，server_crc为None
    if k.server_crc is None:
      print("server crc is None, maybe server not support crc64")
    else:
      print("client_crc:", client_crc, "server_crc:", k.server_crc)
      assert client_crc == k.server_crc, f"client_crc: {client_crc}, server_crc: {k.server_crc}"


# 获取服务端CRC64值，与本地文件计算的CRC64值进行对比
def test_headObject_checkingCrc64():
    k = bucket.new_key(key_name)
    ret = k.set_contents_from_filename(file_path)
    assert ret.status == 200, f"ret.status: {ret.status} (expected 200)"

    k = bucket.get_key(key_name, validate=True)
    client_crc = utils.compute_file_crc64(file_path)
    assert client_crc == k.server_crc, f"client_crc: {client_crc}, k.server_crc: {k.server_crc}"


# copy文件后，获取目标文件的CRC64值，与本地文件计算的CRC64值进行对比
def test_copyObject_checkingCrc64():
    k = bucket.new_key(key_name)
    ret = k.set_contents_from_filename(file_path)
    assert ret.status == 200, f"ret.status: {ret.status} (expected 200)"

    target_key = bucket.copy_key("copy_test2", bucket_name, key_name)
    client_crc = utils.compute_file_crc64(file_path)
    assert client_crc == target_key.server_crc, f"client_crc: {client_crc}, target_key.server_crc: {target_key.server_crc}"


# 这么做，没什么意义。
# copy parts方法没有自动做crc校验（服务端保证），如果确实需要，可以手动计算本地源文件每个part的crc，
# 并通过合并得到最终的crc值，与服务端返回的crc进行对比
def test_multi_copy_checkingCrc64():
    k = bucket.new_key("test_multi")
    ret = k.set_contents_from_filename(file_path)
    assert ret.status == 200, f"ret.status: {ret.status} (expected 200)"

    copy_key = 'test_multi-copy'
    source_size = os.stat(file_path).st_size

    mp = bucket.initiate_multipart_upload(copy_key)

    chunk_size = 100 * 1024
    count = int(math.ceil(source_size * 1.0 / chunk_size * 1.0))

    help_list = []

    for i in range(count):
        start = chunk_size * i
        end = min(start + chunk_size - 1, source_size - 1)
        ret = mp.copy_part_from_key(bucket_name, "test_multi", i + 1, start, end)
        client_crc = utils.compute_file_crc64(file_path, start, end)
        part_info = ks3.multipart.Part()
        part_info.size = end - start + 1
        part_info.part_crc = client_crc
        help_list.append(part_info)
        assert client_crc == ret.server_crc, f"client_crc: {client_crc}, ret.server_crc: {ret.server_crc}"
    ret = mp.complete_upload()
    parts_crc = 0
    crc_obj = utils.Crc64(0)
    for p in help_list:
        parts_crc = crc_obj.combine(parts_crc, int(p.part_crc), p.size)
    assert str(parts_crc) == ret.ChecksumCRC64ECMA, f"parts_crc: {parts_crc}, ret.ChecksumCRC64ECMA: {ret.ChecksumCRC64ECMA}"


########## 手动计算crc，通过请求头校验的方式 ##########
# ks3服务端也提供了crc64校验的请求头，如果不开启自动校验crc的功能，可以通过传入请求头的方式让服务端进行crc的比对
def test_uploadObjectFromFile_checkingCrc64ByHeader():
    k = bucket.new_key(key_name)
    conn.enable_crc = False
    local_crc = utils.compute_file_crc64(file_path)
    crc_header = conn.provider.checksum_crc64ecma_header
    k.set_contents_from_filename(file_path, headers={crc_header: local_crc})


def test_uploadObjectFromString_checkingCrc64ByHeader():
    k = bucket.new_key(key_name)
    myContent = 'some string'
    conn.enable_crc = False
    local_crc = utils.compute_data_crc64(myContent)
    crc_header = conn.provider.checksum_crc64ecma_header
    k.set_contents_from_string(myContent, headers={crc_header: local_crc})


def test_multi_upload_checkingCrc64ByHeader():
    from filechunkio import FileChunkIO
    # 服务端不比对文件整体crc，可以在本地比对
    conn.enable_crc = False
    crc_header = conn.provider.checksum_crc64ecma_header
    crc_obj = utils.Crc64(0)
    final_crc = 0

    source_size = os.stat(file_path).st_size

    mp = bucket.initiate_multipart_upload("test_multi")

    chunk_size = 100 * 1024
    count = int(math.ceil(source_size * 1.0 / chunk_size * 1.0))

    for i in range(count):
        offset = chunk_size * i
        read_size = min(chunk_size, source_size - offset)
        with FileChunkIO(file_path, 'r', offset=offset, bytes=read_size) as fp:
            # 计算单个分片的crc
            local_part_crc = utils.compute_file_crc64(file_path, offset, offset + read_size - 1)
            # 将每一个分片的crc合并，最终将得到文件整体的crc值
            final_crc = crc_obj.combine(final_crc, int(local_part_crc), read_size)
            # 逐个上传分片
            mp.upload_part_from_file(fp, part_num=i + 1, headers={crc_header: local_part_crc})
    ret = mp.complete_upload()

    assert ret.ChecksumCRC64ECMA == str(final_crc), f"ret.ChecksumCRC64ECMA: {ret.ChecksumCRC64ECMA}, final_crc: {final_crc}"
    assert ret and ret.status == 200, f"ret: {ret}, ret.status: {getattr(ret, 'status', None)} (expected 200)"


if __name__ == '__main__':
    # test_downloadObject_checkingCrc64()
    # test_downloadObjectToString_checkingCrc64()
    # test_downloadObjectToFile_checkingCrc64()
    # test_headObject_checkingCrc64()
    # # test_copyObject_checkingCrc64()
    # # test_multi_copy_checkingCrc64()
    # test_uploadObjectFromFile_checkingCrc64ByHeader()
    # test_uploadObjectFromString_checkingCrc64ByHeader()
    # test_multi_upload_checkingCrc64ByHeader()
    test_downloadObjectToString_checkingCrc64()


# -*- coding: utf-8 -*-
from ks3.billing import query_bucket_rank, query_ks3_data

####################公共头#########################

# 金山云主账号 AccessKey 拥有所有API的访问权限，风险很高。强烈建议您创建并使用子账号账号进行 API 访问或日常运维，请登录 https://uc.console.ksyun.com/pro/iam/#/user/list 创建子账号。

# bucket_names 为一个或者多个 bucket 名称，逗号间隔
# start_time 和 end_time 分别为起始时间和结束时间，时间格式为%Y%m%d%H%M%S，比如202503010000
# ks3_product 指的是查询单个或者多个计费项名称，逗号间隔；比如DataSize,RequestsGet；如果不填，则查询所有计费项（除带宽，带宽需单独查询）
# transfer 和 request 分别为可以查询单个或多个统计项的流量/请求次数情况，统计项以逗号分隔。可填参数：Object、Referer、IP、UA，返回TOP200数据
# 详情参考：https://docs.ksyun.com/documents/41328
query_ks3_data('<yourAccessKeyId>', '<yourAccessKeySecret>', bucket_names="<yourBucketName1>,<yourBucketName2>",
               start_time="202503010000", end_time="202503020000", date_type="Day", ks3_product="DataSize,RequestsGet",
               transfer='transfer', request='request')

# bucket_names 为一个或者多个 bucket 名称，逗号间隔
# start_time 和 end_time 分别为起始时间和结束时间，时间格式为%Y%m%d%H%M%S，比如202503010000
# ks3_product 指的是查询单个或者多个计费项名称，逗号间隔；比如DataSize,RequestsGet；如果不填，则查询所有计费项（除带宽，带宽需单独查询）
# number 为TOP排序的 Bucket 数量，取值范围为[1-200]
# 详情参考：https://docs.ksyun.com/documents/43119
query_bucket_rank('<yourAccessKeyId>', '<yourAccessKeySecret>', bucket_names="<yourBucketName1>,<yourBucketName2>",
               start_time="202503010000", end_time="202503020000", date_type="Day", ks3_product="DataSize,RequestsGet",
                  number=10)




# coding:utf-8
import logging
import time
import re
import os

from requests.adapters import DEFAULT_POOLBLOCK
from requests import Request
from urllib3 import poolmanager

from ks3 import utils
from requests.adapters import HTTPAdapter
from ks3.utils import get_default_user_agent

try:
    import http.client as httpcli  # for Python 3
    import urllib.parse as parse
    from urllib import parse as urlparse
except ImportError:
    print('python@2.x no longer supported by ks3sdk@2.x, please refer to ks3sdk@1.x for python@2.x support')

from ks3.auth import canonical_string, add_auth_header, url_encode, encode
from ks3.authV4 import add_auth_header as add_auth_header_v4

logger = logging.getLogger(__name__)


class CallingFormat:
    PATH = 1
    SUBDOMAIN = 2
    VANITY = 3


class AuthingFormat:
    V2 = 1
    V4 = 2


def merge_meta(headers, metadata):
    final_headers = headers.copy()
    for k in list(metadata.keys()):
        final_headers["x-kss-" + "meta-" + k] = metadata[k]

    return final_headers


def query_args_hash_to_string(query_args):
    pairs = []
    for k, v in list(query_args.items()):
        piece = k
        if v is not None:
            piece += "=%s" % parse.quote_plus(str(v).encode('utf-8'))
            # piece += "=%s" % v
        pairs.append(piece)

    return '&'.join(pairs)


# TODO: 删除
def get_object_url(age, bucket="", key="", secret_access_key="", access_key_id="", query_args={}):
    expire = str(int(time.time()) + age)
    headers = {"Date": expire}
    c_string = canonical_string("GET", bucket, key, query_args, headers)
    path = c_string.split("\n")[-1]

    signature = parse.quote_plus(encode(secret_access_key, c_string))
    if "?" in path:
        url = "http://kss.ksyun.com%s&Expires=%s&AccessKeyId=%s&Signature=%s" % \
              (path, expire, access_key_id, signature)
    else:
        url = "http://kss.ksyun.com%s?Expires=%s&AccessKeyId=%s&Signature=%s" % \
              (path, expire, access_key_id, signature)
    return url


def get_proxy_url(scheme, host, port, proxy_username, proxy_password):
    if host is None:
        return None
    if port is None:
        address = host
    else:
        address = '{0}:{1}'.format(host, port)
    if proxy_username is not None and proxy_password is not None:
        proxy_auth = '{0}:{1}@'.format(proxy_username, proxy_password)
    else:
        proxy_auth = ''
    return {scheme: 'http://{0}{1}'.format(proxy_auth, address)}


def do_request(session, method, url, headers=None, data=None, proxy=None, timeout=10):
    headers = headers or {}
    data = data or {}
    proxy = proxy or {}

    req = Request(
        method=method.upper(),
        url=url,
        headers=headers,
        data=data,
    )
    prep = session.prepare_request(req)
    prep.headers.update(headers)
    stream = True
    verify = True
    settings = session.merge_environment_settings(
        prep.url, proxy, stream, verify, None
    )
    send_kwargs = {
        "timeout": timeout,
        "allow_redirects": False,
    }
    send_kwargs.update(settings)
    resp = session.send(prep, **send_kwargs)
    resp = KS3Response(resp)
    return resp


def make_request(server, port, access_key_id, access_key_secret, bucket="", key="", query_args=None, headers=None,
                 data="", metadata=None, method="PUT", calling_format=None, is_secure=False,
                 domain_mode=False, need_auth_header=True, timeout=10, ua_addon='', proxy_host=None,
                 proxy_port=None, proxy_username=None, proxy_password=None, session=None):
    if not headers:
        headers = {}
    # if not query_args:
    #    query_args = {}
    if not metadata:
        metadata = {}

    if bucket and not domain_mode:
        server = calling_format.get_bucket_server(server, bucket)
    path = calling_format.build_path_base(bucket, key)
    # path += "/%s" % url_encode(key)
    # path = path.replace('//', '/%2F')

    if query_args:
        if isinstance(query_args, dict):
            path += "?" + query_args_hash_to_string(query_args)
        else:
            path += "?" + query_args

    host = "%s:%d" % (server, port)

    headers['User-Agent'] = get_default_user_agent() + ' ' + ua_addon
    final_headers = merge_meta(headers, metadata)
    # 修正Content-Length设置逻辑
    # 对于流式上传（chunked transfer），不应设置Content-Length
    # 只有在非chunked传输且data有确定长度时才设置Content-Length
    if method == "PUT" and "Content-Length" not in final_headers and "Transfer-Encoding" not in final_headers and not data:
        final_headers["Content-Length"] = "0"
    if method.upper() == "POST" and "Content-Length" not in final_headers and "Transfer-Encoding" not in final_headers and not data:
        final_headers["Content-Length"] = "0"
    if need_auth_header:
        add_auth_header(access_key_id, access_key_secret, final_headers, method,
                        bucket, key, query_args)
    final_headers['Accept-Encoding'] = 'identity'
    logger.info('send [{method}] request, host: {host}, port: {port}, path: {path}, headers: {headers}'
                .format(method=method, host=host, port=port, path=path, headers=final_headers))

    scheme = 'https' if is_secure else 'http'
    proxy = get_proxy_url(scheme, proxy_host, proxy_port, proxy_username, proxy_password)

    resp = do_request(
        session=session,
        method=method,
        url='{scheme}://{host}{path}'.format(scheme=scheme, host=host, path=path),
        headers=final_headers,
        data=data,
        proxy=proxy,
        timeout=timeout
    )

    logger.info(
        'complete [{method}] request, host: {host}, port: {port}, path: {path}, request_id: {request_id}, '
        'status_code:{status}, headers: {headers}'
        .format(
            method=method,
            host=host,
            port=port,
            path=path,
            request_id=resp.getheader('x-kss-request-id') if resp else '',
            status=resp.status,
            headers=resp.headers
        )
    )
    if 300 <= resp.status < 400:
        loc = resp.getheader('location')
        if loc:
            reg = re.findall(r'http[s]?://(.*?)(:\d+)?/', loc)
            if reg:
                # 如果返回的是bucket style域名，需要提取region域名出来
                if len(reg[0][0].split('.')) == 4:
                    new_server = reg[0][0].split('.', 1)[1]
                else:
                    new_server = reg[0][0]
                loc_parse = urlparse.urlparse(loc)
                if 'Signature' in loc_parse.query:
                    try:
                        resp_temp = do_request(
                            session=session,
                            method=method,
                            url='{scheme}://{host}{path}'.format(
                                scheme=scheme, host=new_server,
                                path=loc_parse.path + '?' + loc_parse.query
                            ),
                            headers=final_headers,
                            data=data,
                            proxy=proxy,
                            timeout=timeout
                        )
                        return resp_temp
                    except Exception as err:
                        print(str(err))
                else:
                    if hasattr(data, 'read'):
                        data.seek(0, os.SEEK_SET)
                        if isinstance(data, utils.FpAdapter):
                            data.reset_crc_process()
                    return make_request(new_server, port, access_key_id, access_key_secret, bucket, key, query_args,
                                        headers, data, metadata, method=method, calling_format=calling_format,
                                        is_secure=is_secure, domain_mode=domain_mode,
                                        need_auth_header=True, timeout=timeout, ua_addon=ua_addon,
                                        proxy_host=proxy_host, proxy_port=proxy_port, session=session)
    return resp


# 发送awsv4的请求
def make_request_v4(access_key_id, access_key_secret, method='', service='', region='', query_args=None, headers={},
                    body="", is_secure=False, timeout=10, inner_api=False):
    inner_string = '.inner' if inner_api else ''
    host = service + inner_string + '.api.ksyun.com'

    if (is_secure):
        connection = httpcli.HTTPSConnection(host)
    else:
        connection = httpcli.HTTPConnection(host)
    connection.timeout = timeout

    path = "/"
    if query_args:
        if isinstance(query_args, dict):
            query_args = query_args_hash_to_string(query_args)
    path += "?" + query_args

    headers = add_auth_header_v4(access_key_id, access_key_secret, region, service, host, method, query_args, body,
                                 headers)

    connection.request(method, path, body, headers)
    resp = connection.getresponse()
    return resp


class KS3Response:
    def __init__(self, resp):
        self.raw_resp = resp  # requests.models.Response
        self.status = resp.status_code
        self.headers = self.msg = resp.raw.headers
        self.reason = resp.reason

    def __bool__(self):
        return self.raw_resp is not None

    def getheader(self, name, default=None):
        return self.headers.get(name, default)

    def getheaders(self):
        return self.headers

    def read(self, amt=None):
        # urllib3.response.HTTPResponse.read()
        return self.raw_resp.raw.read(amt)


def repair_pool_key():
    if 'key_blocksize' in poolmanager.PoolKey._fields:
        return
    try:
        key_fields_tuple = poolmanager._key_fields
        key_list = list(key_fields_tuple)
        key_list.append('key_blocksize')
        patched_key_fields = tuple(key_list)

        import collections
        import functools

        PatchedPoolKey = collections.namedtuple("PatchedPoolKey", patched_key_fields)
        poolmanager.PoolKey = PatchedPoolKey
        poolmanager.key_fn_by_scheme = {
            "http": functools.partial(poolmanager._default_key_normalizer, PatchedPoolKey),
            "https": functools.partial(poolmanager._default_key_normalizer, PatchedPoolKey),
        }
    except AttributeError:
        pass


class KS3HTTPAdapter(HTTPAdapter):
    def __init__(self, block_size=8 * 1024, *args, **kwargs):
        self.block_size = block_size

        repair_pool_key()
        super().__init__(*args, **kwargs)

    def init_poolmanager(self, connections, maxsize, block=DEFAULT_POOLBLOCK, **pool_kwargs):
        if 'key_blocksize' in poolmanager.PoolKey._fields:
            pool_kwargs['blocksize'] = self.block_size
        super().init_poolmanager(connections, maxsize, block=block, **pool_kwargs)



import copy
import logging
from urllib import parse

from ks3 import utils
from ks3.exception import KS3ClientError, S3ResponseError
from ks3.keyfile import KeyFile
from ks3.multipart import Part
from ks3.upload import UploadTask, sse_c_header_keys, UploadResult
from ks3.utils import DEFAULT_PART_SIZE, calculate_string_md5

logger = logging.getLogger(__name__)

# ssec加密copy操作需要 额外 用到的headers
copy_sse_c_header_keys = ['x-kss-copy-source-server-side-encryption-customer-key',
                          'x-kss-copy-source-server-side-encryption-customer-algorithm',
                          'x-kss-copy-source-server-side-encryption-customer-key-MD5']
# copy part需要用到的headers
copy_part_header_keys = ['x-kss-copy-source-if-match', 'x-kss-copy-source-if-none-match',
                         'x-kss-copy-source-if-modified-since', 'x-kss-copy-source-if-unmodified-since',]
copy_part_header_keys.extend(copy_sse_c_header_keys)
copy_part_header_keys.extend(sse_c_header_keys)

# metadata需要忽略的headers
metadata_ignore_header_keys = ['Date', 'Server', 'x-kss-request-id', 'X-Application-Context', 'Connection',
                               'Accept-Ranges',
                               'Content-Length',  # 复制的场景，不设置Content-Length
                               ]


class CopyTask(UploadTask):
    def __init__(
            self, src_key, dest_key,
            executor,
            part_size=DEFAULT_PART_SIZE,
            resumable=False,
            resumable_filename=None,
            copy_src_metadata=True,
            copy_src_tagging=True,
    ):
        """
        src_key: 源key对象，需要包含bucket和connection属性（ks3.key.Key）
        dest_key: 目标key对象，需要包含bucket和connection属性（ks3.key.Key）
        executor: 线程池
        part_size: 分块大小。小于分块大小的文件，使用简单复制；大于分块大小的文件，使用分块复制
        resumable: 是否开启断点续传，默认False。开启断点续传时，如果本地存在有效的断点续传记录，则尝试恢复上传，否则初始化分块上传
        resumable_filename: 断点续传记录文件路径。如果不指定，则使用默认路径（md5[src_bucket/src_key_dest_bucket/dest_key] + '.ks3resume'）
        copy_src_metadata: 是否复制源端的metadata，默认True
        copy_src_tagging: 是否复制源端的tagging，默认True
        """
        super().__init__(dest_key, dest_key.bucket, src_key, executor, part_size, resumable, resumable_filename)
        # 是否使用源端相关信息
        self.copy_src_metadata = copy_src_metadata
        self.copy_src_tagging = copy_src_tagging
        # 源key的metadata
        self._src_metadata = None

    def copy(self, headers=None):
        ur = super().upload(headers)
        # 额外对比一次源端服务端的crc和目标端服务端的crc
        if self.bucket.connection.enable_crc:
            src_crc = self.src_file.server_crc
            dest_crc = ur.crc64ecma
            crc_valid = utils.check_crc(src_crc, dest_crc)
            if not crc_valid:
                if self.resumable:
                    self.record_manager.delete()
                raise KS3ClientError('src_key={0}, dest_key={1}, Inconsistent CRC checksum src_crc: {2}, dest_crc: {3}'
                                     .format(self.src_file.name, self.key.name, src_crc, dest_crc))
        return ur

    def single_transfer(self, headers):
        resp = self.bucket.copy_key(self.key.name, self.src_file.bucket.name, self.src_file.name, headers=headers)
        ur = UploadResult(
            upload_id=None,
            response_metadata=resp.response_metadata,
        )
        ur.etag = resp.etag
        return ur

    def _prepare(self, headers):
        head_headers = {}
        for k in sse_c_header_keys:
            if k in headers:
                head_headers[k] = headers[k]
        self._init_src_info(head_headers)
        self._handle_headers(headers)
        self.part_size = self._adjust_part_size(self.part_size)

    # 如果选择copy源key相关信息，需要将源key的headers覆盖到当前请求的headers中
    def _handle_headers(self, headers):
        if self.copy_src_metadata:
            headers.update(self._src_metadata)
        if self.copy_src_tagging:
            tagging = self.src_file.get_object_tagging()
            if tagging.tagging_set is not None and len(tagging.tagging_set) > 0:
                tagging_value = '&'.join(
                    '{0}={1}'.format(parse.quote(tag.key), parse.quote(tag.value)) for tag in tagging.tagging_set)
                headers['x-kss-tagging'] = tagging_value
        # 如果没有传入任何acl头，则设置和源key一样的acl
        acl_header_keys = ['x-kss-acl', 'x-kss-grant-read', 'x-kss-grant-full-control']
        if not any(k in headers for k in acl_header_keys):
            # 得到源acl并设置
            policy = self.src_file.bucket.get_acl(self.src_file.name)
            read_list = []
            full_control_list = []
            for grant in policy.acl.grants:
                value_format = '{0}="{1}"'
                if grant.email_address is not None:
                    value = value_format.format('emailAddress', grant.email_address)
                elif grant.uri is not None:
                    value = value_format.format('uri', grant.uri)
                else:
                    # 通过请求头设置，只有id会生效
                    value = value_format.format('id', grant.id)
                if grant.permission == 'READ':
                    if 'http://acs.ksyun.com/groups/global/AllUsers' in value:
                        headers['x-kss-acl'] = 'public-read'
                    read_list.append(value)
                elif grant.permission == 'FULL_CONTROL':
                    full_control_list.append(value)
            if 'x-kss-acl' not in headers:
                if read_list:
                    headers['x-kss-grant-read'] = ','.join(read_list)
                if full_control_list:
                    headers['x-kss-grant-full-control'] = ','.join(full_control_list)

    def _put_part(self, mp, part_num, headers):
        start_offset = (part_num - 1) * self.part_size
        remain_bytes = self._file_size - start_offset
        upload_size = min(self.part_size, remain_bytes)
        end_offset = start_offset + upload_size - 1
        resp = mp.copy_part_from_key(self.src_file.bucket.name, self.src_file.name, part_num, start_offset, end_offset,
                                     headers)
        if self.resumable and self.bucket.connection.enable_crc:
            part_info = Part()
            part_info.size = upload_size
            part_info.part_crc = resp.response_metadata.headers[self.bucket.connection.provider.checksum_crc64ecma_header]
            self.record_manager.record.part_crc_infos[part_num] = part_info
            # 保存crc信息
            self.record_manager.save()

    # 初始化源key信息
    def _init_src_info(self, headers=None):
        key, resp = self.src_file.bucket._get_key_internal(self.src_file.name, headers, None)
        if key is None:
            logger.error('Failed to head src key {0} from bucket {1}'
                         .format(self.src_file.name, self.src_file.bucket.name))
            raise S3ResponseError(resp.status, resp.reason, None, request_id=resp.getheader('x-kss-request-id'))
        self.src_file = key
        self._mtime = self.src_file.last_modified
        self._file_size = self.src_file.size
        metadata = {}
        for k in resp.headers:
            if k not in metadata_ignore_header_keys:
                metadata[k] = resp.headers[k]
        self._src_metadata = metadata

    def _gen_resumable_filename(self):
        str_for_md5 = '{0}/{1}_{2}/{3}'.format(
            self.src_file.bucket.name, self.src_file.name, self.bucket.name, self.key.name,
        )
        md5_str = calculate_string_md5(str_for_md5)
        return md5_str + '.ks3resume'

    def _set_header_keys(self):
        super()._set_header_keys()
        self._part_header_keys.extend(copy_part_header_keys)

    def upload(self, headers=None):
        raise NotImplementedError('CopyTask does not support upload method')


class AcrossRegionCopyTask(CopyTask):
    """
    跨region复制，本质上是调用下载接口得到数据流，使用该数据流调用上传接口
    """
    def __init__(
            self, src_key, dest_key,
            executor,
            part_size=DEFAULT_PART_SIZE,
            resumable=False,
            resumable_filename=None,
            copy_src_metadata=True,
            copy_src_tagging=True,
    ):
        super().__init__(
            src_key, dest_key,
            executor,
            part_size,
            resumable,
            resumable_filename,
            copy_src_metadata,
            copy_src_tagging,
        )

    def single_transfer(self, headers):
        key_file_headers = {}
        for k in self._key_file_header_keys:
            if k in headers:
                key_file_headers[k] = headers[k]
        kf = KeyFile(self.src_file, headers=key_file_headers)
        resp = self.key.set_contents_from_file(fp=kf, headers=headers)

        ur = UploadResult(
            upload_id=None,
            response_metadata=resp.response_metadata,
        )
        return ur

    def _put_part(self, mp, part_num, headers):
        # 下载 和 上传分块 的header目前是一样的，可以不筛选
        start_offset = (part_num - 1) * self.part_size
        remain_bytes = self._file_size - start_offset
        upload_size = min(self.part_size, remain_bytes)
        kf = KeyFile(copy.copy(self.src_file), start_offset, upload_size, headers)
        resp = mp.upload_part_from_file(kf, part_num, headers=headers)
        if self.resumable and self.bucket.connection.enable_crc:
            part_info = Part()
            part_info.size = upload_size
            part_info.part_crc = resp.getheader(self.bucket.connection.provider.checksum_crc64ecma_header)
            self.record_manager.record.part_crc_infos[part_num] = part_info
            # 保存crc信息
            self.record_manager.save()

    def _set_header_keys(self):
        UploadTask._set_header_keys(self)
        # 下载传入的header
        self._key_file_header_keys = sse_c_header_keys



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
批量任务 (Batch Operations) 使用示例

该示例演示了如何使用 KS3 批量任务接口，
包括创建、查询、删除和列出批量任务。
"""

import os
from ks3.connection import Connection
from ks3.xmlParsers.batchJob import (
    CreateJobRequest, DescribeJobResult, ListJobsResult,
    KS3PutObjectDataRedundancyTransition, KS3PutObjectAcl, KS3AccessControlList, KS3Grant,
    JobManifest, JobManifestFilter,
    JobManifestLocation, JobManifestSpec, JobOperation, JobReport,
    KS3RestoreObject, KS3DeleteObject
)

# 从环境变量获取配置
access_key = os.environ.get('KS3_IAM_ACCESS_KEY_ID')
secret_key = os.environ.get('KS3_IAM_ACCESS_KEY_SECRET')
bucket_name = os.environ.get('KS3_BUCKET')

# 初始化连接
conn = Connection(access_key, secret_key, host='ks3-cn-beijing.ksyuncs.com')
bucket = conn.get_bucket(bucket_name)


def example_create_redundancy_transition_job():
    """示例1: 创建转换为同城冗余的批量任务"""
    print("\n示例1: 创建转换为同城冗余的批量任务")
    print("-" * 50)

    # 创建清单位置
    filter1 = JobManifestFilter('krn:ksc:ks3:::bucket1', ['Prefix1', 'Prefix2'])
    filter2 = JobManifestFilter('krn:ksc:ks3:::bucket2', ['Prefix1', 'Prefix2'])
    location = JobManifestLocation([filter1, filter2])

    # 创建清单规格
    spec = JobManifestSpec('KS3BatchOperations_Bucket_V1')

    # 创建清单
    manifest = JobManifest(location, spec)

    # 创建操作 - 转换为同城冗余
    operation = JobOperation(data_redundancy_transition=KS3PutObjectDataRedundancyTransition('ZRS'))

    # 创建报告配置
    report = JobReport(
        enabled=True,
        bucket='krn:ksc:ks3:::bucket',
        prefix='job-result',
        report_scope='FailedTasksOnly'
    )

    try:
        result = conn.create_job(
            manifest=manifest,
            operation=operation,
            priority='1',
            description='转换为同城冗余任务',
            report=report,
            client_request_token='unique-token-123'
        )
        print(f"✓ 成功创建批量任务")
        print(f"  任务ID: {result.job_id}")
        print(f"  状态码: {result.status}")
        return result.job_id
    except Exception as e:
        print(f"✗ 创建任务失败: {e}")
        return None


def example_create_restore_job():
    """示例2: 创建批量解冻任务"""
    print("\n示例2: 创建批量解冻任务")
    print("-" * 50)

    # 创建清单
    filter_obj = JobManifestFilter('krn:ksc:ks3:::bucket1', ['archived/'])
    location = JobManifestLocation([filter_obj])
    spec = JobManifestSpec('KS3BatchOperations_Bucket_V1')
    manifest = JobManifest(location, spec)

    # 创建操作 - 批量解冻
    operation = JobOperation(restore_object=KS3RestoreObject(days=3, tier='Standard'))

    # 创建报告配置
    report = JobReport(
        enabled=True,
        bucket='krn:ksc:ks3:::bucket',
        prefix='restore-job-result',
        report_scope='AllTasks'
    )

    try:
        result = conn.create_job(
            manifest=manifest,
            operation=operation,
            priority='1',
            description='批量解冻归档对象',
            report=report
        )
        print(f"✓ 成功创建批量解冻任务")
        print(f"  任务ID: {result.job_id}")
        return result.job_id
    except Exception as e:
        print(f"✗ 创建任务失败: {e}")
        return None


def example_create_acl_job():
    """示例3: 创建批量修改ACL任务（使用预设ACL）"""
    print("\n示例3: 创建批量修改ACL任务（使用预设ACL）")
    print("-" * 50)

    # 创建清单
    filter_obj = JobManifestFilter('krn:ksc:ks3:::bucket1', ['public/'])
    location = JobManifestLocation([filter_obj])
    spec = JobManifestSpec('KS3BatchOperations_Bucket_V1')
    manifest = JobManifest(location, spec)

    # 创建操作 - 批量修改ACL（使用预设ACL）
    operation = JobOperation(put_object_acl=KS3PutObjectAcl('public-read'))

    # 创建报告配置
    report = JobReport(
        enabled=True,
        bucket='krn:ksc:ks3:::bucket',
        prefix='acl-job-report',
        report_scope='FailedTasksOnly'
    )

    try:
        result = conn.create_job(
            manifest=manifest,
            operation=operation,
            priority='0',
            description='批量修改ACL为public-read',
            report=report
        )
        print(f"✓ 成功创建批量修改ACL任务（预设ACL）")
        print(f"  任务ID: {result.job_id}")
        return result.job_id
    except Exception as e:
        print(f"✗ 创建任务失败: {e}")
        return None


def example_create_acl_job_with_grants():
    """示例3-2: 创建批量修改ACL任务（使用自定义ACL）"""
    print("\n示例3-2: 创建批量修改ACL任务（使用自定义ACL）")
    print("-" * 50)

    # 创建清单
    filter_obj = JobManifestFilter('krn:ksc:ks3:::bucket1', ['custom-acl/'])
    location = JobManifestLocation([filter_obj])
    spec = JobManifestSpec('KS3BatchOperations_Bucket_V1')
    manifest = JobManifest(location, spec)

    # 创建授权列表 - 可以配置多个授权
    # KS3Grant 参数: (grantee_id, permission)
    # permission 可选值: FULL_CONTROL, READ
    grants = [
        KS3Grant('123', 'FULL_CONTROL'),  # 用户ID，完全控制权限
        KS3Grant('123', 'READ'),          # 用户ID，只读权限
    ]

    # 创建自定义访问控制列表
    access_control_list = KS3AccessControlList(grants)

    # 创建操作 - 批量修改ACL（使用自定义ACL）
    operation = JobOperation(put_object_acl=KS3PutObjectAcl(access_control_list=access_control_list))

    # 创建报告配置
    report = JobReport(
        enabled=True,
        bucket='krn:ksc:ks3:::bucket',
        prefix='custom-acl-job-report',
        report_scope='FailedTasksOnly'
    )

    try:
        result = conn.create_job(
            manifest=manifest,
            operation=operation,
            priority='5',
            description='批量设置自定义ACL',
            report=report
        )
        print(f"✓ 成功创建批量修改ACL任务（自定义ACL）")
        print(f"  任务ID: {result.job_id}")
        print(f"  授权列表:")
        for grant in grants:
            print(f"    - 授权用户ID: {grant.grantee_id}, 权限: {grant.permission}")
        return result.job_id
    except Exception as e:
        print(f"✗ 创建任务失败: {e}")
        return None


def example_create_delete_job():
    """示例4: 创建批量删除任务"""
    print("\n示例4: 创建批量删除任务")
    print("-" * 50)

    # 创建清单
    filter_obj = JobManifestFilter('krn:ksc:ks3:::bucket1', ['temp/'])
    location = JobManifestLocation([filter_obj])
    spec = JobManifestSpec('KS3BatchOperations_Bucket_V1')
    manifest = JobManifest(location, spec)

    # 创建操作 - 批量删除
    operation = JobOperation(delete_object=KS3DeleteObject())

    # 创建报告配置
    report = JobReport(
        enabled=True,
        bucket='krn:ksc:ks3:::bucket',
        prefix='delete-job-report',
        report_scope='FailedTasksOnly'
    )

    try:
        result = conn.create_job(
            manifest=manifest,
            operation=operation,
            priority='0',
            description='批量删除临时对象',
            report=report
        )
        print(f"✓ 成功创建批量删除任务")
        print(f"  任务ID: {result.job_id}")
        return result.job_id
    except Exception as e:
        print(f"✗ 创建任务失败: {e}")
        return None


def example_describe_job(job_id):
    """示例5: 查询任务详情"""
    print("\n示例5: 查询任务详情")
    print("-" * 50)

    try:
        result = conn.describe_job(job_id)
        print(f"✓ 成功获取任务详情")
        print(f"  任务ID: {result.job_id}")
        print(f"  状态: {result.status}")
        print(f"  创建时间: {result.creation_time}")
        print(f"  描述: {result.description}")
        print(f"  优先级: {result.priority}")
        if result.termination_date:
            print(f"  终止时间: {result.termination_date}")
        if result.progress_summary:
            print(f"  进度: 总任务数={result.progress_summary.total_number_of_tasks}")
            print(f"  成功数量={result.progress_summary.number_of_tasks_succeeded}")
            print(f"  失败数量={result.progress_summary.number_of_tasks_failed}")
    except Exception as e:
        print(f"✗ 查询任务失败: {e}")


def example_delete_job(job_id):
    """示例6: 删除任务"""
    print("\n示例6: 删除任务")
    print("-" * 50)

    try:
        result = conn.delete_job(job_id)
        print(f"✓ 成功删除任务")
        print(f"  任务ID: {job_id}")
        print(f"  状态码: {result.status}")
    except Exception as e:
        print(f"✗ 删除任务失败: {e}")


def example_list_jobs():
    """示例7: 列出任务"""
    print("\n示例7: 列出任务")
    print("-" * 50)

    try:
        result = conn.list_jobs()
        print(f"✓ 成功获取任务列表")
        print(f"  任务数量: {len(result.jobs)}")
        for i, job in enumerate(result.jobs, 1):
            print(f"  任务{i}:")
            print(f"    ID: {job.job_id}")
            print(f"    状态: {job.status}")
            print(f"    描述: {job.description}")
            print(f"    创建时间: {job.creation_time}")
        if result.next_token:
            print(f"  下一页令牌: {result.next_token}")
    except Exception as e:
        print(f"✗ 列出任务失败: {e}")


def example_list_jobs_with_filter():
    """示例8: 列出任务（带过滤）"""
    print("\n示例8: 列出任务（带过滤）")
    print("-" * 50)

    try:
        result = conn.list_jobs(
            next_token=None,
            max_results='10',
            job_statuses='Active'
        )
        print(f"✓ 成功获取任务列表（Active状态）")
        print(f"  任务数量: {len(result.jobs)}")
        for job in result.jobs:
            print(f"  - {job.job_id}: {job.description} ({job.status})")
    except Exception as e:
        print(f"✗ 列出任务失败: {e}")


def example_update_job_priority(job_id):
    """示例9: 更新任务优先级"""
    print("\n示例9: 更新任务优先级")
    print("-" * 50)

    try:
        result = conn.update_job_priority(job_id, '0')
        print(f"✓ 成功更新任务优先级")
        print(f"  任务ID: {result.job_id}")
        print(f"  新优先级: 0")
        print("\n优先级说明:")
        print("  0 - 最高优先级")
        print("  1 - 高优先级")
        print("  10 - 中优先级")
        print("  20 - 低优先级")
    except Exception as e:
        print(f"✗ 更新任务优先级失败: {e}")


def example_xml_format():
    """示例10: 生成的 XML 请求体格式说明"""
    print("\n示例10: 生成的 XML 请求体格式")
    print("-" * 50)

    print("\n创建任务的请求 XML 格式:")
    print("""
<CreateJobRequest>
    <Manifest>
        <Location>
            <Filter>
                <Bucket>krn:ksc:ks3:::bucket1</Bucket>
                <Prefix>Prefix1</Prefix>
                <Prefix>Prefix2</Prefix>
            </Filter>
            <Filter>
                <Bucket>krn:ksc:ks3:::bucket2</Bucket>
                <Prefix>Prefix1</Prefix>
                <Prefix>Prefix2</Prefix>
            </Filter>
        </Location>
        <Spec>
            <Format>KS3BatchOperations_Bucket_V1</Format>
        </Spec>
    </Manifest>
    <Operation>
        <KS3PutObjectDataRedundancyTransition>
            <DataRedundancyType>ZRS</DataRedundancyType>
        </KS3PutObjectDataRedundancyTransition>
    </Operation>
    <Priority>1</Priority>
    <Description>任务描述</Description>
    <Report>
        <Enabled>true</Enabled>
        <Bucket>krn:ksc:ks3:::bucket</Bucket>
        <Prefix>job-result</Prefix>
        <ReportScope>FailedTasksOnly</ReportScope>
    </Report>
    <ClientRequestToken>unique-token</ClientRequestToken>
</CreateJobRequest>
    """)


if __name__ == '__main__':
    print("=" * 50)
    print("批量任务 (Batch Operations) 使用示例")
    print("=" * 50)

    # 注意: 以下示例需要有效的 KRN (Ksyun Resource Name) 和实际的 bucket 才能正常工作
    # 运行前请确保环境变量已正确配置

    # 取消下面的注释来运行示例
    # job_id = example_create_redundancy_transition_job()
    # if job_id:
    #     example_describe_job(job_id)
    #     example_update_job_priority(job_id)
    #     example_delete_job(job_id)

    # job_id2 = example_create_restore_job()
    # if job_id2:
    #     example_describe_job(job_id2)

    # example_create_acl_job()
    # example_create_delete_job()

    # example_list_jobs()
    # example_list_jobs_with_filter()

    # 显示 XML 格式说明（无需实际环境）
    example_xml_format()

    print("\n" + "=" * 50)
    print("示例运行完毕")
    print("=" * 50)


import json
from typing import List

from ks3.responseResult import ResponseMetadata

from ks3.exception import KS3ClientError

class BucketDecompressPolicy:
    def __init__(self, rules: List['Rule'] = None, *args, **kwargs):
        self.rules = []
        for rule in rules:
            self.add_rule(rule)

        self.response_metadata = ResponseMetadata(**kwargs)

    def get_rules(self) -> List['Rule']:
        return self.rules

    def set_rules(self, rules: List['Rule']):
        for rule in rules:
            self.add_rule(rule)

    def add_rule(self, rule: 'Rule'):
        rule.validate()
        self.rules.append(rule)

    def to_json(self) -> str:
        # 转json时 不要带response_metadata，具体实现如下
        try:
            return json.dumps(self, default=lambda o: {k: v for k, v in o.__dict__.items() if k != 'response_metadata'})
        except Exception as e:
            raise KS3ClientError("failed to serialize decompressPolicy to json text", e)

    @staticmethod
    def from_json(json_text: str) -> 'BucketDecompressPolicy':
        try:
            data = json.loads(json_text)
            rules = [Rule(**rule) for rule in data.get('rules', [])]
            return BucketDecompressPolicy(rules)
        except Exception as e:
            raise KS3ClientError("failed to deserialize decompressPolicy from json text", e)

class OverwritePolicy:
    SKIP_IF_EXISTS = 0
    OVERWRITE_IF_EXISTS = 1

class PathPrefixPolicy:
    ADD_FILENAME = 0
    JUST_PATH_PREFIX = 1

class Rule:
    MAXIMUM_ALLOWED_ID_LENGTH = 256

    def __init__(self, id: str = None, events: str = None, prefix: str = None, suffix: List[str] = None,
                 overwrite: int = OverwritePolicy.SKIP_IF_EXISTS, callback: str = None, callback_format: str = None,
                 path_prefix: str = None, path_prefix_replaced: int = PathPrefixPolicy.ADD_FILENAME,
                 policy_type: str = "decompress"):
        self.id = id
        self.events = events
        self.prefix = prefix
        self.suffix = suffix if suffix is not None else []
        self.overwrite = overwrite
        self.callback = callback
        self.callback_format = callback_format
        self.path_prefix = path_prefix
        self.path_prefix_replaced = path_prefix_replaced
        self.policy_type = policy_type

    def validate(self):
        self.validate_rule_id()
        self.validate_events()
        self.validate_callback()
        self.validate_path_prefix()
        self.validate_policy_type()

    def validate_rule_id(self):
        if not self.id:
            raise KS3ClientError("missing rule id")
        if len(self.id) > self.MAXIMUM_ALLOWED_ID_LENGTH:
            raise KS3ClientError(f"rule id length must be between 1 and {self.MAXIMUM_ALLOWED_ID_LENGTH}")

    def validate_events(self):
        if not self.events:
            raise KS3ClientError("missing events")

    def validate_callback(self):
        # 如果callback不为空，则必须要有callback_format
        if self.callback:
            self.validate_callback_format()

    def validate_callback_format(self):
        if not self.callback_format:
            raise KS3ClientError("missing callback_format")

    def validate_path_prefix(self):
        if self.path_prefix and not self.path_prefix.endswith("/"):
            raise KS3ClientError("pathPrefix must end with '/' if not empty")

    def validate_policy_type(self):
        if not self.policy_type:
            raise KS3ClientError("missing policyType")

    def __repr__(self):
        return f"Rule(id={self.id}, events={self.events}, prefix={self.prefix}, suffix={self.suffix}, " \
               f"overwrite={self.overwrite}, callback={self.callback}, callback_format={self.callback_format}, " \
               f"path_prefix={self.path_prefix}, path_prefix_replaced={self.path_prefix_replaced}, " \
               f"policy_type={self.policy_type})"

# 示例用法
if __name__ == "__main__":
    rule = Rule(id="rule1", events="ObjectCreated:*", prefix="prefix1", suffix=[".zip"], path_prefix="path/")
    rule.validate()
    policy = bucketDecompressPolicy()
    policy.add_rule(rule)
    json_text = policy.to_json()
    print(json_text)
    new_policy = bucketDecompressPolicy.from_json(json_text)
    print(new_policy.get_rules())



from ks3.responseResult import ResponseMetadata


class Tagging(object):
    def __init__(self, tagging_set=None, *args, **kwargs):
        if tagging_set is None:
            self.tagging_set = []
        else:
            self.tagging_set = tagging_set

        self.response_metadata = ResponseMetadata(**kwargs)

    def __repr__(self):
        s = "tagging "
        for tag in self.tagging_set:
            s += tag.key + ':' + (lambda x: ['', tag.value][tag.value is not None])(tag.value) + ' '
        return s

    def startElement(self, name, attrs, connection):
        if name == 'Tag':
            self.tagging_set.append(Tag())
            return self.tagging_set[-1]
        else:
            return None

    def endElement(self, name, value, connection):
        setattr(self, name, value)

    def to_xml(self):
        s = '<?xml version="1.0" encoding="UTF-8"?>'
        s += '<Tagging>'
        if self.tagging_set is not None:
            s += '<TagSet>'
            for tag in self.tagging_set:
                s += tag.to_xml()
            s += '</TagSet>'
        s += '</Tagging>'
        return s


class Tag(object):
    def __init__(self, key=None, value=None):
        self.key = key
        self.value = value

    def startElement(self, name, attrs, connection):
        return None

    def endElement(self, name, value, connection):
        if name == 'Key':
            self.key = value
        elif name == 'Value':
            self.value = value
        else:
            setattr(self, name, value)

    def to_xml(self):
        s = '<Tag>'
        if self.key is not None:
            s += '<Key>%s</Key>' % self.key
        if self.value is not None:
            s += '<Value>%s</Value>' % self.value
        s += '</Tag>'
        return s
